from MC_class import *
import pickle
import time

## If debug model, set as 1
debug = 0

## Root of labelled vessel
OOI = ROI('../image/marked.tif')

## hyperparameters
particles_total = 150
init_frame_idx = 7300
end_frame_idx = 7803

# vessel type, 0 negative direction, 1 positive direction
vessel_dir = 0
max_velocity = 50
min_velocity = 8
# MC search accuracy
velocity_error = 2
particles_intensity_search_step = 2
# Parameter of meanshift algorithm
MeanShift_bandwidth = 5
min_particles_in_cluster = 10
min_component_weights = 0.15


time_start = time.clock()
## initialize MC_frames class for each frame
MC_frames = []
for frame_idx in range(init_frame_idx, end_frame_idx+1, 1):
    MC_frames.append(MC_frame(frame_idx))

## compute temporal average frame
Temporal_avg = []
print('copying files to the result folder')
for frame_idx in range(init_frame_idx, end_frame_idx+1, 1):
    Im_name = '../image/Registered Images/' + str(frame_idx) + '.tif'
    Im_raw = cv2.imread(Im_name, cv2.IMREAD_GRAYSCALE)
    Im_raw = cv2.GaussianBlur(Im_raw, (5, 5), 0.5)
    Temporal_avg.append(Im_raw)
    ## copy all the frame to the result folder
    cv2.imwrite('../image/results/' + str(frame_idx) + '.tif', Im_raw)
print('copy finishes')
Temporal_avg = np.asarray(Temporal_avg)
Temporal_avg = np.mean(Temporal_avg, 0).astype(np.uint8)
if debug:
    cv2.imshow('a', Temporal_avg)
    cv2.waitKey()




## Traverse each frame as init frame
for start_frame_idx in range(1, end_frame_idx-init_frame_idx+1):
    print('**********************************')
    print('Init frame:' + str(start_frame_idx + init_frame_idx))
    ## Read each frame current frame and former frame, and get variance map
    former_frame = Im_frame()
    former_frame.initialization(start_frame_idx + init_frame_idx - 1,
                                 raw_image_root='../image/Registered Images/', temporal_avg=Temporal_avg)
    current_frame = Im_frame()
    current_frame.initialization(start_frame_idx + init_frame_idx,
                                 raw_image_root='../image/Registered Images/', temporal_avg=Temporal_avg)
    current_frame.Im_postprocessing(former_frame)
    init_frame = copy.deepcopy(current_frame)

    ## MC particles initialization in current frame, 'running_particles_clusters' is a list of particle_clusters
    running_particles_cluster_init = MC_frames[start_frame_idx].particle_init(particles_total, OOI)
    # show init results
    if debug:
        current_frame.draw_particles_all(running_particles_cluster_init,
                                                window_name='Init: frame' + str(start_frame_idx+init_frame_idx))

    ## particle intensity move
    running_particles_cluster_init.intensity_move(Frame=current_frame, search_step=particles_intensity_search_step)
    eff_check = running_particles_cluster_init.weight_compute(Frame=current_frame)
    if eff_check == False:
        print('discard: weights of initialized particles are low')
        if not debug:
            init_frame.draw_particle_clusters([], window_name=str(start_frame_idx+init_frame_idx), radius=7, save_flag=1)
        for j in range(MC_frames[start_frame_idx].particle_cluster_list.__len__()):
            print('particle_idx:')
            print(MC_frames[start_frame_idx].particle_cluster_list[j].cluster_idx)
            print('velocity:')
            print(MC_frames[start_frame_idx].particle_cluster_list[j].compute_velocity_average())
            print('pos:')
            print(MC_frames[start_frame_idx].particle_cluster_list[j].compute_pos_average())
        continue

    # show intensity_move results
    if debug:
        current_frame.draw_particles_all(running_particles_cluster_init,
                                                window_name='Init: frame' + str(start_frame_idx+init_frame_idx) + ' intensity_move')

    ## resampling
    running_particles_cluster_init.weight_compute_norm(resample_flag=False)
    for _ in range(10):
        running_particles_cluster_init.resampling(particles_total)
        running_particles_cluster_init.weight_compute_norm(resample_flag=True)
    # show resampling results
    if debug:
        current_frame.draw_particles_all(running_particles_cluster_init,
                                                window_name='Init: frame' + str(start_frame_idx+init_frame_idx) + ' resampling')

    ## reclustering,
    running_particles_cluster_init.weight_compute(Frame=current_frame)
    running_particles_clusters = running_particles_cluster_init.clustering(
        Frame=current_frame, MeanShift_bandwidth=MeanShift_bandwidth, min_particles_in_cluster=min_particles_in_cluster, min_component_weights=min_component_weights)
    # add particle clusters into frame
    running_particles_clusters = MC_frames[start_frame_idx].add_cluster_from_clusters(running_particles_clusters, MC_frames[start_frame_idx-1].compute_maximum_cluster_idx(), cts_flag=False)
    # show reclustering results
    if debug:
        current_frame.draw_particle_clusters(running_particles_clusters,
                                                window_name='Init: frame' + str(start_frame_idx+init_frame_idx) + ' reclustering',
                                             radius=7)

    ## move the particles
    for cts_frame_idx in range(start_frame_idx+1, end_frame_idx - init_frame_idx + 1):
        # read new frame
        print('cts frame:' + str(cts_frame_idx + init_frame_idx))

        former_frame = copy.deepcopy(current_frame)
        current_frame = Im_frame()
        current_frame.initialization(cts_frame_idx + init_frame_idx,
                                     raw_image_root='../image/Registered Images/', temporal_avg=Temporal_avg)
        current_frame.Im_postprocessing(former_frame)

        # velocity_move
        running_particles_clusters_temp = []
        for i in range(running_particles_clusters.__len__()):
            eff_check = running_particles_clusters[i].velocity_move(OOI, min_velocity, max_velocity,
                                                    vessel_dir, velocity_error, current_frame,
                                                                    search_step=particles_intensity_search_step)
            if eff_check:
                running_particles_clusters_temp.append(running_particles_clusters[i])
        running_particles_clusters = copy.deepcopy(running_particles_clusters_temp)
        if running_particles_clusters.__len__() == 0:
            print('velocity move fails')
            break

        # show velocity_move results
        if debug:
            current_frame.draw_particle_clusters_all(running_particles_clusters,
                                                    window_name='cts: frame' + str(
                                                        cts_frame_idx + init_frame_idx) + ' velocity move',
                                                    radius=1)

        ## weight compute
        running_particles_clusters_temp = []
        for i in range(running_particles_clusters.__len__()):
            eff_check = running_particles_clusters[i].weight_compute(Frame=current_frame)
            if eff_check:
                running_particles_clusters_temp.append(running_particles_clusters[i])
        running_particles_clusters = copy.deepcopy(running_particles_clusters_temp)

        if running_particles_clusters.__len__() == 0:
            print('discard: weights of particles are low')
            break
        if debug:
            current_frame.draw_particle_clusters_all(running_particles_clusters,
                                                    window_name='cts: frame' + str(
                                                        cts_frame_idx + init_frame_idx) + ' discard low weights',
                                                    radius=1)


        ## resampling
        for i in range(running_particles_clusters.__len__()):
            running_particles_clusters[i].weight_compute_norm(resample_flag=False)
            for _ in range(10):
                running_particles_clusters[i].resampling(particles_total)
                running_particles_clusters[i].weight_compute_norm(resample_flag=True)

        # show resampling result
        if debug:
            current_frame.draw_particle_clusters_all(running_particles_clusters,
                                                    window_name='cts: frame' + str(
                                                        cts_frame_idx + init_frame_idx) + ' resampling',
                                                    radius=1)

        ## reclustering
        running_particles_clusters_temp = []
        for i in range(running_particles_clusters.__len__()):
            running_particles_clusters[i].weight_compute(Frame=current_frame)
            running_particles_single_temp = running_particles_clusters[i].clustering(
                Frame=current_frame, MeanShift_bandwidth=MeanShift_bandwidth, min_particles_in_cluster=min_particles_in_cluster,
                min_component_weights=min_component_weights)

            if running_particles_clusters.__len__() == 1:
                running_particles_single_temp[0].cluster_idx = running_particles_clusters[i].cluster_idx
                running_particles_clusters_temp.append(running_particles_single_temp[0])
        running_particles_clusters = copy.deepcopy(running_particles_clusters_temp)

        if running_particles_clusters.__len__() == 0:
            print('fails because of clustering split')
            break

        running_particles_clusters_temp = []
        for i in range(running_particles_clusters.__len__()):
            eff_check = running_particles_clusters[i].weight_compute(Frame=current_frame)
            if eff_check:
                running_particles_clusters_temp.append(running_particles_clusters[i])
        running_particles_clusters = copy.deepcopy(running_particles_clusters_temp)


        running_particles_clusters = clusters_discard(running_particles_clusters, min_component_weights, particles_total)
        if running_particles_clusters.__len__() == 0:
            print('discard: weights of particles are low after reclustering')
            break

        # show reclustering results
        if debug:
            current_frame.draw_particle_clusters(running_particles_clusters,
                                                 window_name='cts: frame' + str(
                                                     cts_frame_idx + init_frame_idx) + ' reclustering',
                                                 radius=7)


        running_particles_clusters = MC_frames[cts_frame_idx].add_cluster_from_clusters(running_particles_clusters, MC_frames[cts_frame_idx-1].compute_maximum_cluster_idx(), cts_flag=True)

    # save result
    if not debug:
        effective_particle_idx, color_map = init_frame.draw_particle_clusters(MC_frames[start_frame_idx].particle_cluster_list,
                                            window_name=str(start_frame_idx+init_frame_idx),
                                            radius=7, save_flag=1)

        former_result = cv2.imread('../image/results/' + str(start_frame_idx+init_frame_idx-1) + '.tif', cv2.IMREAD_COLOR)
        for j in range(MC_frames[start_frame_idx-1].particle_cluster_list.__len__()):
            for jj in range(effective_particle_idx.__len__()):
                if MC_frames[start_frame_idx-1].particle_cluster_list[j].cluster_idx == effective_particle_idx[jj]:
                    pos_avg_x, pos_avg_y = MC_frames[start_frame_idx-1].particle_cluster_list[j].compute_pos_average()
                    idx = (effective_particle_idx[jj] - 1) % 12
                    color = color_map[idx]
                    cv2.circle(former_result, center=(np.uint32(pos_avg_x), np.uint32(pos_avg_y)),
                               radius=7, color=color, thickness=1)
                    break
        cv2.imwrite('../image/results/' + str(start_frame_idx+init_frame_idx-1) + '.tif', former_result)




    for j in range(MC_frames[start_frame_idx].particle_cluster_list.__len__()):
        print('particle_idx:')
        print(MC_frames[start_frame_idx].particle_cluster_list[j].cluster_idx)
        print('velocity:')
        print(MC_frames[start_frame_idx].particle_cluster_list[j].compute_velocity_average())
        print('pos:')
        print(MC_frames[start_frame_idx].particle_cluster_list[j].compute_pos_average())

time_elapsed = (time.clock() - time_start)
print(time_elapsed)
with open('../image/results/MC_frames.pkl', 'wb') as f1:
    pickle.dump(MC_frames, f1)